/*
 * Copyright (C) 2009-2012 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "uart_soft.h"

#include "config.h"
#include "avrutils.h"

#include <avr/pgmspace.h>


// adjust the bit length a bit for all the instructions in between the SET/RESET calls
#define BIT_LENGTH_US (1000000./SOFT_UART_BAUD_RATE - (16./F_CPU))

void uart_send(uint8_t data) {
  if( data == '\n' )
    uart_send('\r');

#if SOFT_UART_DISABLE_INTERRUPTS
  const auto sreg = SREG;
  cli();
#endif

// This should be assembly, ah well. It's easier this way and works for 9600 Baud.
RESET(SOFT_UART_OUT);
_delay_us(BIT_LENGTH_US);
for(uint8_t i = 8; i > 0; --i) {
  if(data & 1)
    SET(SOFT_UART_OUT);
  else
    RESET(SOFT_UART_OUT);
  data >>= 1;
  _delay_us(BIT_LENGTH_US);
}
SET(SOFT_UART_OUT);
_delay_us(BIT_LENGTH_US);

#if SOFT_UART_DISABLE_INTERRUPTS
  SREG = sreg;
#endif
}

void uart_soft_init(void) {
  SET(SOFT_UART_OUT);
  SET_OUTPUT(SOFT_UART_OUT);
}

void uart_puts_P(const char* str) {
  char c;
  str--;
  while((c = pgm_read_byte((void*)(++str)))) {
    uart_send(c);
  }
  uart_send('\n');
}
